SystemOrganization addCategory: #GODSpreadsheet!

GODData subclass: #SSSpreadsheetData
	instanceVariableNames: 'sheets'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODSpreadsheet'!

!SSSpreadsheetData commentStamp: 'jjs 10/18/2014 17:56' prior: 0!
Creates a spreadsheet with a set of sheets!

!SSSpreadsheetData class methodsFor: 'private' stamp: 'foa 11/19/2014 15:14'!
fromCSV: filePath
	"initialize a SSSpreadsheetData from specified CSV"
	| file |
	file := FileStream fileNamed: filePath.
	^self fromCSVStream: file withSeparator: ','.! !

!SSSpreadsheetData class methodsFor: 'private' stamp: 'foa 11/19/2014 15:14'!
fromCSVStream: csvStream withSeparator: separator
	"Returns a SSSpreadsheetData from given CSV file stream. Separator indicates character to use as cell separator in csv."

	| line row columns ssdata sheet count|
	
	ssdata := SSSpreadsheetData new.
	sheet := ssdata createSheetWithName: 'default'.
	
	[csvStream atEnd] whileFalse:
		[line := csvStream nextLine.
		columns := line findTokens: separator escapedBy: '"'.
		row := sheet createRow.
		count := 1.
		columns do: [:item | row createCellAtColumnIndex: count withValue: item. count := count + 1]
		].
	^ssdata! !

!SSSpreadsheetData class methodsFor: 'initializing' stamp: 'dhk 10/31/2014 23:27'!
fromCollection: aCollection
	"initialize a SSSpreadsheetData with a single SSSheet and fills it with aCollection"
	|data sheet|
	data := SSSpreadsheetData new.
	sheet := data createSheetWithName: 'default'.
	sheet fillFromCollection: aCollection.
	
	^data
	! !

!SSSpreadsheetData class methodsFor: 'initializing' stamp: 'foa 10/31/2014 23:20'!
fromDictionary: aDictionary
	"initialize a SSSpreadsheetData with a single SSSheet and fills it with aDictionary"
	
	| ssdata sheet | 
	ssdata := SSSpreadsheetData new.
	sheet := ssdata createSheetWithName: 'default'.
	sheet fillFromDictionary: aDictionary.
	
	^ssdata! !

!SSSpreadsheetData class methodsFor: 'initializing' stamp: 'foa 11/1/2014 18:18'!
fromFile: filePath
	"initialize a SSSpreadsheetData from specified path"
	|extension|
	extension := (filePath subStrings: '.') last asUppercase.
	extension = 'CSV'
		ifTrue: [^self fromCSV: filePath].
	extension = 'ODS'
		ifTrue: [^self fromODS: filePath].
	extension = 'XLSX'
		ifTrue: [^self fromXLSX: filePath].! !

!SSSpreadsheetData class methodsFor: 'private' stamp: 'foa 10/18/2014 18:16'!
fromODS: filePath
	"initialize a SSSpreadsheetData from specified ODS"
	
	| odsFile contentsXML data root ssdata sheet row colIndex rowsRepeat columnsRepeat |
	
	odsFile := ZipArchive new readFrom: filePath.
	contentsXML := odsFile member: 'content.xml'.
	data := XMLDOMParser parseDocumentFrom: contentsXML contents readStream.
	root := data elementAt: #'office:document-content'.
	
	ssdata := SSSpreadsheetData new.
	root tagsNamed: #'table:table' do: [ :table |
		sheet := ssdata createSheetWithName: (table attributeAt: 'table:name').
		table tagsNamed: #'table:table-row' do: [ :trow |
			rowsRepeat := trow attributeAt: #'table:number-rows-repeated'.
			rowsRepeat isNil ifTrue: [ rowsRepeat := 1 ] ifFalse: [ rowsRepeat := rowsRepeat asNumber ].
			rowsRepeat timesRepeat: [
				row := sheet createRow.
				colIndex := 1.
				trow tagsNamed: #'table:table-cell' do: [ :cell |
					columnsRepeat := cell attributeAt: #'table:number-columns-repeated'.
					columnsRepeat isNil ifTrue: [ columnsRepeat := 1 ] ifFalse: [ columnsRepeat := columnsRepeat asNumber ].
					columnsRepeat timesRepeat: [
						row createCellAtColumnIndex: colIndex withValue: (SSSpreadsheetData getODSCellValue: cell).
						colIndex := colIndex + 1.
						].
					].
				].
			].
		].
	^ssdata.! !

!SSSpreadsheetData class methodsFor: 'private' stamp: 'dhk 11/2/2014 17:50'!
fromXLSX: filePath
	"initialize a SSSpreadsheetData from specified XLSX"

	| xlsxFile stringsXML stringsRoot sharedStrings workbookXML workbookRoot ssdata sheet sheetXML sheetRoot row | 
	xlsxFile := ZipArchive new readFrom: filePath.
	
	"1) store the strings in a ordered collection"
	stringsXML := xlsxFile member: 'xl/sharedStrings.xml'.
	stringsRoot := (XMLDOMParser parseDocumentFrom: stringsXML contents readStream) elementAt: 'sst'.
	sharedStrings := OrderedCollection new.
	stringsRoot tagsNamed: #'t' do: [ :t |
		sharedStrings add: (t elementsAndContents first characterData).
		].
		
	"2) get the name and number of sheets in the file"
	workbookXML := xlsxFile member: 'xl/workbook.xml'.
	workbookRoot := (XMLDOMParser parseDocumentFrom: workbookXML contents readStream) elementAt: 'workbook'.
	ssdata := SSSpreadsheetData new.
	workbookRoot tagsNamed: #'sheet' do: [ :sheetTag |
		sheet := ssdata createSheetWithName: (sheetTag attributeAt: 'name').
		
		"3) for each sheet, load its contents"
		sheetXML := xlsxFile member: ('xl/worksheets/sheet' , (sheetTag attributeAt: 'sheetId') , '.xml').
		sheetRoot := (XMLDOMParser parseDocumentFrom: sheetXML contents readStream) elementAt: 'worksheet'.
		
		"3.A) first, iterate through rows"
		sheetRoot tagsNamed: #'row' do: [ :rowTag |
			row := sheet createRowAtIndex: ((rowTag attributeAt: 'r') asNumber).
			
			"3.B) then, iterate over the cells in this row"
			rowTag tagsNamed: #'c' do: [ :cellTag |
				row createCellAtColumn: ((cellTag attributeAt: 'r') onlyLetters) withValue: (SSSpreadsheetData getXLSXCellValue: cellTag withStrings: sharedStrings).
				].
			].
		].
	
	^ssdata! !

!SSSpreadsheetData class methodsFor: 'utility' stamp: 'dhk 11/2/2014 19:09'!
getDateFromXLSXValue: value
	"Converts the date value from a XLSX spreadsheet, which is the number of days since 01/01/1900, to its equivalent string representation day/month/year."
	
	| dayInit v |
	"It seems that it has a bug... It shows 2 days ahead than we expect... So..."
	dayInit := (Date year: 1900 month: 01 day: 01) addDays: -2.
	
	v := (dayInit addDays: value).
	
	^v printFormat: #(1 2 3 $/ 1 1 2). "format like this: 31/10/1999"! !

!SSSpreadsheetData class methodsFor: 'utility' stamp: 'foa 11/15/2014 16:30'!
getODSCellValue: aCell
	"Gets the value of the given cell (from a ODS spreadsheet)"

	| internalData |
	internalData := aCell firstTagNamed: #'text:p'.
	internalData isNil ifTrue: [ ^nil ].
	
	^internalData elementsAndContents first characterData.! !

!SSSpreadsheetData class methodsFor: 'utility' stamp: 'foa 11/1/2014 21:30'!
getTimeFromXLSXValue: value
	"Converts the time value from a XLSX spreadsheet, which is a percentage between 0 and 1, to its equivalent string representation hour:minute of the day."

	| time hours minutes minStr |
	time := 24 * value.
	hours := time asInteger.
	minutes := ((time fractionPart) * 60) rounded.
	minStr := minutes asString.
	minStr size = 1 ifTrue: [ minStr := '0' , minStr ].
	^(hours asString) , ':' , minStr
	! !

!SSSpreadsheetData class methodsFor: 'utility' stamp: 'dhk 11/2/2014 19:08'!
getXLSXCellValue: cellTag withStrings: sharedStrings
	"Gets the value of the given cell tag (from a XLSX spreadsheet), with the given shared strings collection."

	| internalValue type subType date time | 
	internalValue := (cellTag firstTagNamed: #'v') elementsAndContents first characterData.
	
 	type := cellTag attributeAt: #'t'.
	type isNil 
		ifTrue: [
			subType := cellTag attributeAt: #'s'.
			subType isNil ifFalse: [
				subType := subType asNumber.
				subType = 1 ifTrue: [ "Date value"
					internalValue := SSSpreadsheetData getDateFromXLSXValue: (internalValue asNumber).
					].
				subType = 2 ifTrue: [ "Time value"
					internalValue := SSSpreadsheetData getTimeFromXLSXValue: (internalValue asNumber).
					].
				"subType=3 should be Percentage, but in this case we'll store the same value (1.2, as stored in the XLSX, instead of 120%)"
				subType = 4 ifTrue: [ "Date+time value"
					internalValue := internalValue asNumber.
					date := SSSpreadsheetData getDateFromXLSXValue: (internalValue integerPart).
					time := SSSpreadsheetData getTimeFromXLSXValue: (internalValue fractionPart).
					internalValue := time , ' ' , date.
					].
				].
			^internalValue
			]
		ifFalse: [
			type = 's' ifTrue: [
				"This cell is a string in the shared strings collection, indexed by internalValue."
				^sharedStrings at: ((internalValue asNumber) + 1).
				].
			].
	^internalValue! !

!SSSpreadsheetData class methodsFor: 'private' stamp: 'dhk 11/15/2014 17:46'!
skelManifest
^'<?xml version="1.0" encoding="UTF-8"?>
<manifest:manifest xmlns:manifest="urn:oasis:names:tc:opendocument:xmlns:manifest:1.0" manifest:version="1.2">
 <manifest:file-entry manifest:full-path="/" manifest:version="1.2" manifest:media-type="application/vnd.oasis.opendocument.spreadsheet"/>
 <manifest:file-entry manifest:full-path="content.xml" manifest:media-type="text/xml"/>
</manifest:manifest>'! !

!SSSpreadsheetData methodsFor: 'adding' stamp: 'dhk 10/29/2014 22:16'!
createSheetWithName: name
	"creates a sheet with name 'name'"
	|sheet|
	sheet := SSSheet new: name.
	sheet parent: self.
	sheets add: sheet.
	sheet index: sheets size.
	^sheet! !

!SSSpreadsheetData methodsFor: 'accessing' stamp: 'dhk 10/17/2014 18:17'!
getSheet: sheetNumber
	"returns the sheet with the sheetNumber index"
	
	^sheets at: sheetNumber.! !

!SSSpreadsheetData methodsFor: 'accessing' stamp: 'dhk 10/17/2014 18:23'!
getSheetWithName: sheetName
	"returns the sheet with name sheetName"
	sheets do:
		[:sheet | (sheet name = sheetName) ifTrue: [^sheet]].
	^nil! !

!SSSpreadsheetData methodsFor: 'initialize' stamp: 'jjs 10/18/2014 17:55'!
initialize
	"creates a new empty set of sheets and returns the set of sheets"
	sheets := OrderedCollection new.
	^self! !

!SSSpreadsheetData methodsFor: 'accessing' stamp: 'dhk 9/27/2014 17:23'!
numberOfSheets
	"returns how many sheets exists in this GODSpreadsheetData"

	^sheets size! !

!SSSpreadsheetData methodsFor: 'accessing' stamp: 'jjs 10/18/2014 17:39'!
renameSheet: sheetName toName: aName
	"renames a sheet with name sheetName to the name aName"
	|sheet|
	sheet := (self getSheetWithName: sheetName).
	sheet changeName: aName.! !

!SSSpreadsheetData methodsFor: 'file out' stamp: 'dhk 11/15/2014 17:42'!
toFile: filePath
	"save this SSSpreadsheetData to specified path. Returns false if there is an error"

	|extension|
	extension := (filePath subStrings: '.') last asUppercase.
	extension = 'ODS'
		ifTrue: [^self toODS: filePath].

	^true! !

!SSSpreadsheetData methodsFor: 'private' stamp: 'dhk 11/15/2014 17:57'!
toODS: filePath
	"save this SSSpreadsheetData to specified path. Returns false if there is an error"

	| tempdir metadir zip |
	
	tempdir := FileDirectory default / '_GODSS_exporting'.
	
	tempdir assureExistence; recursiveDelete; assureExistence.
	
	metadir := tempdir / 'META-INF'.
	
	metadir assureExistence.
	
	metadir fileNamed: 'manifest.xml' do: [ :stream | stream nextPutAll: SSSpreadsheetData skelManifest ].
	tempdir fileNamed: 'content.xml' do: [ :stream | self writeODSContentToStream: stream ].
	
	zip := ZipArchive new.
	zip addDirectory: metadir fullName as: 'META-INF'.
	zip addFile: (metadir fullNameFor: 'manifest.xml') as: ('META-INF', FileDirectory slash, 'manifest.xml').
	zip addFile: (tempdir fullNameFor: 'content.xml') as: 'content.xml'.
	
	zip writeToFileNamed: filePath.
	zip close.
	
	tempdir recursiveDelete.
	
	^true! !

!SSSpreadsheetData methodsFor: 'private' stamp: 'foa 11/15/2014 16:21'!
writeODSContentToStream: odsStream
	"Write in the given stream this spreadsheet contents, as a ODS contents.xml format. Used to save this spreadsheet to a ODS file."

	odsStream nextPutAll: '<?xml version="1.0" encoding="UTF-8"?><office:document-content xmlns:office="urn:oasis:names:tc:opendocument:xmlns:office:1.0" xmlns:style="urn:oasis:names:tc:opendocument:xmlns:style:1.0" xmlns:text="urn:oasis:names:tc:opendocument:xmlns:text:1.0" xmlns:table="urn:oasis:names:tc:opendocument:xmlns:table:1.0" xmlns:draw="urn:oasis:names:tc:opendocument:xmlns:drawing:1.0" xmlns:fo="urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0" xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:meta="urn:oasis:names:tc:opendocument:xmlns:meta:1.0" xmlns:number="urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0" xmlns:presentation="urn:oasis:names:tc:opendocument:xmlns:presentation:1.0" xmlns:svg="urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0" xmlns:chart="urn:oasis:names:tc:opendocument:xmlns:chart:1.0" xmlns:dr3d="urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0" xmlns:math="http://www.w3.org/1998/Math/MathML" xmlns:form="urn:oasis:names:tc:opendocument:xmlns:form:1.0" xmlns:script="urn:oasis:names:tc:opendocument:xmlns:script:1.0" xmlns:ooo="http://openoffice.org/2004/office" xmlns:ooow="http://openoffice.org/2004/writer" xmlns:oooc="http://openoffice.org/2004/calc" xmlns:dom="http://www.w3.org/2001/xml-events" xmlns:xforms="http://www.w3.org/2002/xforms" xmlns:xsd="http://www.w3.org/2001/XMLSchema" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:rpt="http://openoffice.org/2005/report" xmlns:of="urn:oasis:names:tc:opendocument:xmlns:of:1.2" xmlns:xhtml="http://www.w3.org/1999/xhtml" xmlns:grddl="http://www.w3.org/2003/g/data-view#" xmlns:tableooo="http://openoffice.org/2009/table" xmlns:drawooo="http://openoffice.org/2010/draw" xmlns:calcext="urn:org:documentfoundation:names:experimental:calc:xmlns:calcext:1.0" xmlns:field="urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0" xmlns:formx="urn:openoffice:names:experimental:ooxml-odf-interop:xmlns:form:1.0" xmlns:css3t="http://www.w3.org/TR/css3-text/" office:version="1.2">'.
	odsStream nextPutAll: '<office:scripts/><office:font-face-decls><style:font-face style:name="Arial" svg:font-family="Arial" style:font-family-generic="swiss" style:font-pitch="variable"/><style:font-face style:name="DejaVu Sans" svg:font-family="&apos;DejaVu Sans&apos;" style:font-family-generic="system" style:font-pitch="variable"/><style:font-face style:name="Lohit Hindi" svg:font-family="&apos;Lohit Hindi&apos;" style:font-family-generic="system" style:font-pitch="variable"/></office:font-face-decls>'.
	odsStream nextPutAll: '<office:automatic-styles><style:style style:name="co1" style:family="table-column"><style:table-column-properties fo:break-before="auto" style:column-width="3.000cm"/></style:style><style:style style:name="ro1" style:family="table-row"><style:table-row-properties style:row-height="0.500cm" fo:break-before="auto" style:use-optimal-row-height="true"/></style:style><style:style style:name="ta1" style:family="table" style:master-page-name="Default"><style:table-properties table:display="true" style:writing-mode="lr-tb"/></style:style></office:automatic-styles>'.
	
	odsStream nextPutAll: '<office:body><office:spreadsheet>'.
	sheets do: [ :sheet | sheet writeODSTableToStream: odsStream ].
	odsStream nextPutAll: '<table:named-expressions/></office:spreadsheet></office:body></office:document-content>'.
	^nil! !

Object subclass: #SSCell
	instanceVariableNames: 'value parent column'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODSpreadsheet'!

!SSCell commentStamp: 'jjs 10/18/2014 17:46' prior: 0!
creates an cell object with a parent row, column index and cell's value!

!SSCell class methodsFor: 'initialize' stamp: 'dhk 10/17/2014 16:45'!
newWithParent: aParent atColumn: aColumnName

	^((self new) parent: aParent) column: aColumnName.! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/26/2014 15:44'!
column
	"Return the column name of this cell.
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."

	^column! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/29/2014 17:51'!
column: aColumn
	"Set the column name of this cell.
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."

	column := (aColumn asUppercase).! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/17/2014 16:22'!
getColumnIndex
	"Returns the column index of this cell."

	^SSSheet columnNameToIndex: column! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/17/2014 16:47'!
getParentRow
	"Returns the Row which this cell belongs to."

	^self parent.! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/17/2014 16:23'!
getRowIndex
	"Returns the row index of this cell. "

	^parent index.! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/17/2014 16:47'!
parent
	"Returns the Row which this cell belongs to."

	^parent.! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/17/2014 16:06'!
parent: aParent
	"set the parent of this cell"

	parent := aParent! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/2/2014 21:40'!
value
	"returns the value of this cell"

	^value! !

!SSCell methodsFor: 'accessing' stamp: 'dhk 10/2/2014 21:40'!
value: aValue
	"set the value of this cell"

	value := aValue! !

!SSCell methodsFor: 'private' stamp: 'dhk 11/15/2014 17:54'!
writeODSCellToStream: odsStream
	"Write this cell to the stream in ODS format."

	value isNil 
		ifTrue: [ odsStream nextPutAll: '<table:table-cell/>'. ]
		ifFalse: [ odsStream nextPutAll: '<table:table-cell office:value-type="string"><text:p>', value, '</text:p></table:table-cell>'. ].

	^nil! !

Object subclass: #SSRow
	instanceVariableNames: 'cells parent index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODSpreadsheet'!

!SSRow commentStamp: 'jjs 10/18/2014 17:49' prior: 0!
Creates and row object with a set of cells, parent sheet and row index!

!SSRow methodsFor: 'adding' stamp: 'dhk 10/26/2014 15:46'!
addCell: aCell atColumn: columnName
	"add aCell at the column with name columnName
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."
	self addCell: aCell atColumnIndex: (SSSheet columnNameToIndex: columnName).
	
	
	
	! !

!SSRow methodsFor: 'adding' stamp: 'dhk 10/26/2014 15:46'!
addCell: aCell atColumnIndex: columnIndex
	"add a cell in this row at the specified column index (an integer)"
	[((cells size)) < columnIndex]
		whileTrue:
		[cells add: (SSCell newWithParent: self atColumn: (SSSheet columnIndexToName: cells size + 1))].
	cells at: columnIndex put: aCell.! !

!SSRow methodsFor: 'deleting' stamp: 'dhk 11/1/2014 10:50'!
clearCell: columnName
	"clear the content of this cell at the column with name columnName
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."
	self clearCellAtIndex: (SSSheet columnNameToIndex: columnName).! !

!SSRow methodsFor: 'deleting' stamp: 'dhk 11/1/2014 10:48'!
clearCellAtIndex: columnIndex
	"clear the content of this cell at specified column index (an integer)"
	(cells at: columnIndex) value: nil.! !

!SSRow methodsFor: 'initialize' stamp: 'dhk 11/6/2014 14:50'!
copyValuesFrom: otherRow
	"initialize this row as a copy of otherRow."
	
	otherRow do: [ :cell | self createCellAtColumn: (cell column) withValue: (cell value) ]! !

!SSRow methodsFor: 'adding' stamp: 'dhk 10/26/2014 15:46'!
createCellAtColumn: columnName withValue: aValue
	"create cell at the column with name columnName and value aValue
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."
	|aCell|
	aCell := (SSCell newWithParent: self atColumn: columnName) value: aValue.
	self addCell: aCell atColumn: columnName.
	^aCell
	! !

!SSRow methodsFor: 'adding' stamp: 'dhk 10/26/2014 15:46'!
createCellAtColumnIndex: columnIndex withValue: aValue
	"create cell at the column index columnIndex (an integer) with value aValue"
	^self createCellAtColumn: (SSSheet columnIndexToName: columnIndex) withValue: aValue
	
	! !

!SSRow methodsFor: 'accessing' stamp: 'dhk 11/1/2014 00:10'!
do: aBlock
"do aBlock for each cell in this row"

cells do: aBlock.! !

!SSRow methodsFor: 'accessing' stamp: 'dhk 10/26/2014 15:45'!
getCell: columnName
	"get the cell in this row at the column with name columnName
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."

	^self getCellAtIndex: (SSSheet columnNameToIndex: columnName).! !

!SSRow methodsFor: 'accessing' stamp: 'dhk 10/26/2014 15:45'!
getCellAtIndex: columnIndex
	"get the cell in this row at columnIndex (an integer)"

	^cells at: columnIndex.! !

!SSRow methodsFor: 'accessing' stamp: 'dhk 11/6/2014 14:28'!
getNumberOfCells
"return the number of cells in this row"

^cells size.! !

!SSRow methodsFor: 'accessing' stamp: 'dhk 10/26/2014 15:45'!
index
	"Returns the index (an integer) of this row in the parent sheet."

	^index.! !

!SSRow methodsFor: 'accessing' stamp: 'dhk 10/17/2014 16:16'!
index: aIndex
	"set the index of this cell"
	index := aIndex! !

!SSRow methodsFor: 'initialize' stamp: 'jjs 10/18/2014 17:28'!
initialize
      "initialize this instance with an empty cell collection"
	cells := OrderedCollection new.! !

!SSRow methodsFor: 'accessing' stamp: 'dhk 10/17/2014 16:52'!
parent
	"Returns the sheet which this row belongs to."

	^parent.! !

!SSRow methodsFor: 'accessing' stamp: 'jjs 10/18/2014 17:28'!
parent: aParent
     "Sets the sheet which this row belongs to."

	parent := aParent.! !

!SSRow methodsFor: 'private' stamp: 'dhk 11/15/2014 17:41'!
writeODSRowToStream: odsStream
	"Write the contents of this row to the given stream in ODS format."
	| count maxColumns |

	odsStream nextPutAll: '<table:table-row table:style-name="ro1">'.
	cells do: [ :cell | cell writeODSCellToStream: odsStream ].
	count := cells size.
	maxColumns := parent getMaximumNumberOfColumns.
	[ count <= maxColumns ] whileTrue:
		[
			odsStream nextPutAll: '<table:table-cell/>'.
			count := count + 1.
		].
	odsStream nextPutAll: '</table:table-row>' .
	^nil! !

Object subclass: #SSSheet
	instanceVariableNames: 'sheetName rows parent index'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'GODSpreadsheet'!

!SSSheet commentStamp: 'jjs 10/18/2014 17:51' prior: 0!
Creates an spreadsheet object with a spreadsheetData parent, a sheet name, a set of rows and a sheet index !

!SSSheet class methodsFor: 'utility' stamp: 'foa 10/17/2014 17:51'!
columnIndexToName: aIndex
	"Utility method to convert a column index value to a column name."
	
	| rest subIndex |
	aIndex <= 26
		ifTrue:	[^(aIndex+64) asCharacter asString].
	rest := (aIndex - 1) \\ 26.
	subIndex := (aIndex - 1) // 26.
	^(self columnIndexToName: subIndex), ((rest + 65) asCharacter asString)! !

!SSSheet class methodsFor: 'utility' stamp: 'dhk 10/29/2014 17:50'!
columnNameToIndex: columnName
	"Utility method to convert a column name value to a column index."
     "index = 0
	l = list(name)
	index = ord(l[0]) - 64
	for i in xrange(1, len(l)):
		index *= 26
		index += ord(l[i]) - 64
	return index"
	
	| index aName | 
	aName := (columnName asUppercase).
	index := (aName first asInteger) - 64.
	aName allButFirstDo: [ :char |
		index := (index * 26) + ((char asInteger) - 64)].
	^index! !

!SSSheet class methodsFor: 'instance creation' stamp: 'dhk 9/27/2014 16:35'!
new: aName
	^(self new) changeName: aName! !

!SSSheet methodsFor: 'private' stamp: 'dhk 9/27/2014 16:39'!
changeName: aName
	"change the name of this sheet"

	sheetName := aName! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/29/2014 18:05'!
createCellAtRow: rowIndex atColumn: columnName withValue: aValue
	"creates a cell at this specified address with value aValue and returns it
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."
	| row cell |
	rows size < rowIndex
		ifTrue: [row := (self createRowAtIndex: rowIndex)]
		ifFalse:[row := rows at: rowIndex].
	cell := row createCellAtColumn: columnName withValue: aValue.
	^cell! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/29/2014 18:07'!
createCellAtRow: rowIndex atColumnIndex: columnIndex withValue: aValue
	"creates a cell at this specified address with value aValue and returns it"
	^self createCellAtRow: rowIndex atColumn: (SSSheet columnIndexToName: columnIndex) withValue: aValue.! !

!SSSheet methodsFor: 'adding' stamp: 'jjs 10/18/2014 17:32'!
createRow
	"creates a new empty row after the last row in this sheet and returns it"
	|row|
	row := ((SSRow new) parent: self).
	rows add: row.
	row index: (rows size).
	^row.! !

!SSSheet methodsFor: 'adding' stamp: 'dhk 10/29/2014 22:10'!
createRowAtIndex: aIndex
	"creates a new empty row at a specified index in this sheet.
	This method creates empty rows if needed"
	
	[((rows size) + 1) < aIndex]
		whileTrue:
			[self createRow].
	^self createRow! !

!SSSheet methodsFor: 'private' stamp: 'dhk 11/6/2014 20:18'!
decideHowToMergeRows: aRow and: otherRow
	| cell1 cell2 |
	cell1 := aRow getCell: 'A'.
	cell2 := otherRow getCell: 'A'.
	
	((cell1 isNil) or: (cell1 value isNil)) ifTrue: [^2].
	((cell2 isNil) or: (cell2 value isNil)) ifTrue: [^1].
	(cell1 value < cell2 value) ifTrue: [^1].
	(cell1 value > cell2 value) ifTrue: [^2].
	^3.! !

!SSSheet methodsFor: 'deleting' stamp: 'dhk 10/17/2014 19:28'!
deleteRow: aIndex
	"delete row at index aIndex"
	
	rows removeAt: aIndex.! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 11/6/2014 14:28'!
do: aBlock
"do aBlock for each row in this sheet"

rows do: aBlock.! !

!SSSheet methodsFor: 'file handling' stamp: 'dhk 11/1/2014 00:29'!
exportToCSV: filePath
	"Export this sheet to a csv file. Cells containing comma ( , ) will be escaped with double quotes"
	| file |
	file := FileStream fileNamed: filePath.
	rows do: [:row |
		row do: [ :cell |
			(cell value includesSubString: ',')
			ifTrue: 
				[file nextPut: $"; nextPutAll: cell value; nextPut: $";nextPut: $,. ]
			ifFalse:
				[file nextPutAll: cell value; nextPut: $,. ].
			].
		file position: file position -1; cr.
		].

	file close.! !

!SSSheet methodsFor: 'initialize' stamp: 'dhk 10/31/2014 23:29'!
fillFromCollection: aCollection
"Fill this sheet with values from aCollection.
aCollection must be a collection of collections"
| row colIndex |
aCollection do:  [ :lineC |
	row := self createRow.
	colIndex := 1.
	lineC do: [ :item |
		row createCellAtColumnIndex: colIndex withValue: item.
		colIndex := colIndex + 1.
		].
	].! !

!SSSheet methodsFor: 'initialize' stamp: 'foa 10/31/2014 23:17'!
fillFromDictionary: aDictionary
"Fill this sheet with values from aDictionary. The dictionary must be a simple one, without collections. Each key-value pair will be a row,
 with the key being in column A and value being in column B."

| row | 
aDictionary keysDo: [ :key |
	row := self createRow.
	row createCellAtColumn: 'A' withValue: key.
	row createCellAtColumn: 'B' withValue: (aDictionary at: key).
	]! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/26/2014 15:48'!
getCellAtRow: rowIndex atColumn: columnName
	"returns the cell at this specified address
	We expect column names as used in common office spreadsheet editing applications. Columns start from A to Z, then AA to AZ, BA and so on..."

	^self getCellAtRow: rowIndex atColumnIndex: (SSSheet columnNameToIndex: columnName).! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/17/2014 17:22'!
getCellAtRow: rowIndex atColumnIndex: columnIndex
	"returns the cell at this specified address"

	^(rows at: rowIndex) getCellAtIndex: columnIndex.! !

!SSSheet methodsFor: 'private' stamp: 'JJS 11/15/2014 17:18'!
getMaximumNumberOfColumns
	"Returns the maximum number of columns amongst our rows."
	| maxNumber |
	maxNumber := 0.
      rows do: [ :row| 
		row getNumberOfCells > maxNumber 
	             ifTrue: [
				maxNumber := row getNumberOfCells.
			].
		 ].
	^maxNumber! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 11/6/2014 14:29'!
getNumberOfRows
	"return the number of rows in this sheet"
	^rows size.! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/17/2014 17:20'!
getRow: rowIndex
	"returns the row with the specified index"

	^rows at: rowIndex! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/29/2014 22:16'!
index
	"return the index of this sheet in the SpreadsheetData"

	^index.! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/29/2014 22:16'!
index: aIndex
	"set the index of this sheet in the SpreadsheetData"

	index := aIndex.! !

!SSSheet methodsFor: 'initialize' stamp: 'dhk 10/17/2014 17:18'!
initialize
	rows := OrderedCollection new.! !

!SSSheet methodsFor: 'merging' stamp: 'dhk 11/6/2014 20:19'!
mergeWith: otherSheet from: fromIndex
	"merge with otherSheet and return a new SSSpreadsheetData with the resulting sheet. This method DOESN'T change the original sheets"
	| newData sheet nlines1 nlines2 i1 i2 iResult row1 row2 rowResult whatToDo columnIndex value1 |
	newData := SSSpreadsheetData new.
	sheet := newData createSheetWithName: 'default'.
	nlines1 := self getNumberOfRows.
	nlines2 := otherSheet getNumberOfRows.
	
	i1 := fromIndex.
	i2 := fromIndex.
	iResult := fromIndex.
	
	[i1 <= nlines1 and: i2 <= nlines2]
		whileTrue:
			[
			"decide what to do: (1) add row from self; (2) add row from otherSheet; (3) merge columns from these rows and add the result."
			row1 := self getRow: i1.
			row2 := otherSheet getRow: i2.
			whatToDo := self decideHowToMergeRows: row1 and: row2.
			whatToDo = 1 ifTrue: [
				(sheet createRow) copyValuesFrom: row1.
				i1 := i1 + 1.
				].
			whatToDo = 2 ifTrue: [
				(sheet createRow) copyValuesFrom: row2.
				i2 := i2 + 1.
				].
			whatToDo = 3 ifTrue: [
					"here we merge columns from these rows"
					rowResult := sheet createRow.
					columnIndex := 1.
					[ columnIndex <= row1 getNumberOfCells ]
						whileTrue:
							[ value1 := (row1 getCellAtIndex: columnIndex) value.
							value1 notNil
								ifTrue: [rowResult createCellAtColumnIndex: columnIndex withValue: value1]
								ifFalse: [rowResult createCellAtColumnIndex: columnIndex withValue: (row2 getCellAtIndex: columnIndex) value].
							columnIndex := columnIndex + 1.
							].
					[ columnIndex <= row2 getNumberOfCells ]
						whileTrue:
							[rowResult createCellAtColumnIndex: columnIndex withValue: (row2 getCellAtIndex: columnIndex) value.
							columnIndex := columnIndex + 1].
					i1 := i1 + 1.
					i2 := i2 + 1.
				]. 
			].
	
	[i1 <= nlines1]
		whileTrue:
			[
				(sheet createRow) copyValuesFrom: (self getRow: i1).
				i1 := i1 + 1.
			].
		
	[i2 <= nlines2]
		whileTrue:
			[
				(sheet createRow) copyValuesFrom: (otherSheet getRow: i2).
				i2 := i2 + 1.
			].
	^newData.! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/2/2014 21:39'!
name
	"returns the name of this sheet"

	^sheetName! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/17/2014 17:19'!
parent
	"return the SSSpreadsheetData which this sheet belongs to"

	^parent! !

!SSSheet methodsFor: 'accessing' stamp: 'dhk 10/17/2014 17:19'!
parent: aParent
	"set the SSSpreadsheetData which this sheet belongs to"

	parent := aParent! !

!SSSheet methodsFor: 'sorting' stamp: 'dhk 11/1/2014 12:13'!
sort
	"sort (using in-place quick sort) all rows by column A"	
	^self sortFrom: 1.! !

!SSSheet methodsFor: 'sorting' stamp: 'dhk 11/1/2014 12:06'!
sortFrom: fromIndex
	"sort rows (using in-place quick sort) from fromIndex to the last row by column A"
	^self sortFrom: fromIndex to: rows size.! !

!SSSheet methodsFor: 'sorting' stamp: 'dhk 11/1/2014 12:06'!
sortFrom: fromIndex to: toIndex
	"sort rows (using in-place quick sort) from fromIndex to toIndex by column A"
	^self sortFrom: fromIndex to: toIndex by: 
	[:r1 :r2 | (r1 getCell: 'A') isNil ifTrue: [false] 
			ifFalse: [ (r2 getCell: 'A') isNil ifTrue: [true]
				ifFalse: [ (r1 getCell: 'A') value isNil ifTrue: [false]
					ifFalse: [ (r2 getCell: 'A') value isNil ifTrue: [true]
						ifFalse: [ (r1 getCell: 'A') value < (r2 getCell: 'A') value ]
						]
					]
				]
			]! !

!SSSheet methodsFor: 'sorting' stamp: 'dhk 11/1/2014 12:06'!
sortFrom: fromIndex to: toIndex by: sortBlock
	"sort rows (using in-place quick sort) from fromIndex to toIndex according to sortBlock"
	rows := OrderedCollection newFrom: ((rows asArray) quickSortFrom: fromIndex to: toIndex by: sortBlock)! !

!SSSheet methodsFor: 'private' stamp: 'foa 11/15/2014 16:22'!
writeODSTableToStream: odsStream
	"Write this sheet to a stream, in ODS format."

	odsStream nextPutAll: ('<table:table table:name="' , sheetName, '" table:style-name="ta1">').
	rows do: [ :row |  row writeODSRowToStream: odsStream. ].
	odsStream nextPutAll: '</table:table>'.
	^nil.! !
